import os
import glob
import argparse
from Bio import SeqIO
from itertools import groupby

def cigar_splitter(cigar):

    # get the position of letters
    letter_pos_list = []
    n = 0
    for each_element in cigar:
        if (each_element.isalpha() is True) or (each_element == '='):
            letter_pos_list.append(n)
        n += 1

    # split cigar
    index = 0
    cigar_splitted = []
    while index <= len(letter_pos_list) - 1:
        if index == 0:
            cigar_splitted.append(cigar[:(letter_pos_list[index] + 1)])
        else:
            cigar_splitted.append(cigar[(letter_pos_list[index - 1] + 1):(letter_pos_list[index] + 1)])
        index += 1

    return cigar_splitted


def check_both_ends_clipping(cigar_splitted):

    both_ends_clipping = False
    if len(cigar_splitted) >= 3:
        if (cigar_splitted[0][-1] in ['S', 's']) and (cigar_splitted[-1][-1] in ['S', 's']):
            both_ends_clipping = True

    return both_ends_clipping


def get_cigar_stats(cigar_splitted):

    # aligned_len: M I X =
    # clipping_len: S
    # mismatch_len: X I D
    # mismatch_pct = mismatch_len / aligned_len
    # aligned_pct  = aligned_len  / (aligned_len + clipping_len)
    # clipping_pct = clipping_len / (aligned_len + clipping_len)

    aligned_len = 0
    clipping_len = 0
    mismatch_len = 0
    for each_part in cigar_splitted:
        each_part_len = int(each_part[:-1])
        each_part_cate = each_part[-1]

        # get aligned_len
        if each_part_cate in {'M', 'm', 'I', 'i', 'X', 'x', '='}:
            aligned_len += each_part_len

        # get clipping_len
        if each_part_cate in ['S', 's', 'H', 'h']:
            clipping_len += each_part_len

        # get mismatch_len
        if each_part_cate in {'I', 'i', 'X', 'x', 'D', 'd'}:
            mismatch_len += each_part_len

    aligned_pct  = float("{0:.2f}".format(aligned_len * 100 / (aligned_len + clipping_len)))
    clipping_pct = float("{0:.2f}".format(clipping_len * 100 / (aligned_len + clipping_len)))
    mismatch_pct = float("{0:.2f}".format(mismatch_len * 100 / (aligned_len)))

    return aligned_len, aligned_pct, clipping_len, clipping_pct, mismatch_pct



sam_file    = '/Users/songweizhi/Desktop/test/map_to.fa.no.secondary.sam'
sam_file    = '/Users/songweizhi/Desktop/test/test.sam'
bin_folder  = '/Users/songweizhi/Desktop/test/metawrap_30_90_bins_with_completed_renamed'
bin_ext     = 'fa'
output_file = '/Users/songweizhi/Desktop/test/abund.txt'

bin_file_re = '%s/*%s' % (bin_folder, bin_ext)
bin_file_list = [os.path.basename(file_name) for file_name in glob.glob(bin_file_re)]

bin_size_dict = dict()
ctg_2_bin_dict = dict()
counted_ctg_set = set()
ctgs_found_in_multiple_mags = set()
for each_bin in bin_file_list:
    pwd_each_bin = '%s/%s' % (bin_folder, each_bin)
    bin_size_dict[each_bin] = 0
    for seq in SeqIO.parse(pwd_each_bin, 'fasta'):
        ctg_2_bin_dict[seq.id] = each_bin
        bin_size_dict[each_bin] += len(seq.seq)
        if seq.id not in counted_ctg_set:
            counted_ctg_set.add(seq.id)
        else:
            ctgs_found_in_multiple_mags.add(seq.id)


read_in_sam = set()
reads2gnm_dict = dict()
for each_read in open(sam_file):
    if not each_read.startswith('@'):
        each_read_split = each_read.strip().split('\t')
        read_id = each_read_split[0]
        ref_id = each_read_split[2]
        ref_pos = int(each_read_split[3])
        cigar = each_read_split[5]
        read_in_sam.add(read_id)
        if ref_id == '*':
            gnm_id = 'unmapped'
        else:
            gnm_id = ctg_2_bin_dict.get(ref_id, 'unbinned')

        if read_id not in reads2gnm_dict:
            reads2gnm_dict[read_id] = dict()
            reads2gnm_dict[read_id][gnm_id] = [cigar]
        else:
            if gnm_id not in reads2gnm_dict[read_id]:
                reads2gnm_dict[read_id][gnm_id] = [cigar]
            else:
                reads2gnm_dict[read_id][gnm_id].append(cigar)

gnm_to_aligned_read_dict = dict()
for each_read in reads2gnm_dict:
    mapped_gnms = reads2gnm_dict[each_read]

    if len(mapped_gnms) == 1:
        for mapped_gnm in mapped_gnms:
            if mapped_gnm not in gnm_to_aligned_read_dict:
                gnm_to_aligned_read_dict[mapped_gnm] = {each_read}
            else:
                gnm_to_aligned_read_dict[mapped_gnm].add(each_read)
    elif (len(mapped_gnms) == 2) and ('unbinned' in mapped_gnms):
        for mapped_gnm in mapped_gnms:
            if mapped_gnm != 'unbinned':
                if mapped_gnm not in gnm_to_aligned_read_dict:
                    gnm_to_aligned_read_dict[mapped_gnm] = {each_read}
                else:
                    gnm_to_aligned_read_dict[mapped_gnm].add(each_read)
    else:
        gnm_with_longest_aln_len = ''
        longest_aln_len = 0
        for mapped_gnm in mapped_gnms:
            if mapped_gnm != 'unbinned':
                for each_cigar in mapped_gnms[mapped_gnm]:
                    cigar_splitted = cigar_splitter(each_cigar)
                    aln_len, aln_pct, clp_len, clp_pct, mismatch_pct = get_cigar_stats(cigar_splitted)
                    if aln_len > longest_aln_len:
                        longest_aln_len = aln_len
                        gnm_with_longest_aln_len = mapped_gnm

        if gnm_with_longest_aln_len not in gnm_to_aligned_read_dict:
            gnm_to_aligned_read_dict[gnm_with_longest_aln_len] = {each_read}
        else:
            gnm_to_aligned_read_dict[gnm_with_longest_aln_len].add(each_read)













import os
import argparse


PhyloBiAssoc_usage = '''
=========== PhyloBiAssoc example commands ===========

BioSAK PhyloBiAssoc -t demo.tre -d demo.txt

# Note: the header for the first two columns has to be "ID" and "cate".

# It will perform:
# 1) binaryPGLMM test if phylosig p-value <= 0.05 (significant phylogenetic signal)
# 2) chi-squared test if phylosig p-value > 0.05  (no phylogenetic signal)
# 3) do nothing if phylosig returns NaN (might due to the same value across all genomes)

# https://www.rdocumentation.org/packages/ape/versions/5.7-1/topics/binaryPGLMM

=====================================================
'''


def PhyloBiAssoc(args):

    tree_file        = args['t']
    data_file        = args['d']
    pwd_current_script  = os.path.realpath(__file__)
    current_script_path = '/'.join(pwd_current_script.split('/')[:-1])
    PhyloBiAssoc_R   = '%s/PhyloBiAssoc.R' % current_script_path
    PhyloBiAssoc_cmd = 'Rscript %s -t %s -d %s' % (PhyloBiAssoc_R, tree_file, data_file)
    os.system(PhyloBiAssoc_cmd)


if __name__ == "__main__":
    PhyloBiAssoc_parser = argparse.ArgumentParser()
    PhyloBiAssoc_parser.add_argument('-t', required=True, help='tree file')
    PhyloBiAssoc_parser.add_argument('-d', required=True, help='data file')
    args = vars(PhyloBiAssoc_parser.parse_args())
    PhyloBiAssoc(args)




##########################################################################################

num_1 = float("{0:.2f}".format(123.456789))  # 123.45
num_2 = float("{0:.3f}".format(123.456789))  # 123.456


# try to int, if not float
num_str = '123.456789'
try:
    pwy_cpl = int(num_str)
except:
    pwy_cpl = float(num_str)


def norm_num_list(num_list, norm_by, decimal_to_keep):
    num_list_after_norm = []
    for each_num in num_list:
        each_num_after_norm = float(("{0:.%sf}" % decimal_to_keep).format(each_num/norm_by))
        num_list_after_norm.append(each_num_after_norm)
    return num_list_after_norm


##########################################################################################

def numericable_str(my_string):

    is_numericable_str = False
    try:
        my_numeric_value = int(my_string)
        is_numericable_str = True
    except ValueError:
        try:
            my_numeric_value = float(my_string)
            is_numericable_str = True
        except ValueError:
            is_numericable_str = False

    return is_numericable_str


############################################## subset_df ##############################################

def subset_df(file_in, rows_to_keep, cols_to_keep, sep_symbol, row_name_pos, column_name_pos, file_out):

    df = pd.read_csv(file_in, sep=sep_symbol, header=column_name_pos, index_col=row_name_pos)

    if len(rows_to_keep) == 0:
        if len(cols_to_keep) == 0:
            subset_df = df.loc[:, :]
        else:
            subset_df = df.loc[:, cols_to_keep]
    else:
        if len(cols_to_keep) == 0:
            subset_df = df.loc[rows_to_keep, :]
        else:
            subset_df = df.loc[rows_to_keep, cols_to_keep]

    subset_df.to_csv(file_out, sep=sep_symbol)

rows_to_keep    = set()
cols_to_keep    = set()
sep_symbol      = '\t'
row_name_pos    = 0
column_name_pos = 0


################################################### Distance matrix ####################################################

'''
Format_of_symmetric_distance_matrix_file (tab separated)

    A   B   C
A   1   2   3
B   4   5   6
C   7   8   9
'''

import pandas as pd

dm_file = 'demo_files/symmetric_distance_matrix.tab'

# read in symmetric distance matrix file
df = pd.read_table(dm_file, header=0, index_col=0)

# df to arrary
df_in_arrary = df.to_numpy()
# [[1 2 3]
#  [4 5 6]
#  [7 8 9]]

# get column name list
column_name_list = list(df.columns)
# ['A', 'B', 'C']

# get row name list
row_name_list = list(df.index)
# ['A', 'B', 'C']

################################### cluster number list ##################################

def cluster(data, maxgap, min_size):

    data.sort()
    groups = [[data[0]]]
    for x in data[1:]:
        if abs(x - groups[-1][-1]) <= maxgap:
            groups[-1].append(x)
        else:
            groups.append([x])

    groups_qualified = []
    for each_cluster in groups:
        if len(each_cluster) >= min_size:
            groups_qualified.append(each_cluster)

    return groups_qualified


################################### get best number of clusters ##################################

import kmedoids
import pandas as pd
from sklearn.metrics import silhouette_score


def get_num_of_clusters(dm_file):

    df = pd.read_table(dm_file, header=0, index_col=0)
    df_in_arrary = df.to_numpy()

    max_score = -1
    best_cluster_num = None
    for cluster_num in range(2, df.shape[0]):

        # perform clustering
        clusters = kmedoids.fasterpam(df, cluster_num)
        label_list_1_based = [int((i + 1)) for i in clusters.labels]

        # calculate silhouette score
        score = silhouette_score(df_in_arrary, label_list_1_based)
        if score > max_score:
            max_score = score
            best_cluster_num = cluster_num

    if max_score < 0.55:
        best_cluster_num = 1

    return best_cluster_num, max_score


dm_file = 'demo_files/edge_20_wd39_dm.tab'
cluster_num, score = get_num_of_clusters(dm_file)
print('Best\t%s\t%s' % (cluster_num, score))
# Best	2	0.8067079676284064


##################################### filter dataframe columns #####################################


def filter_col(df_in, min_value, min_value_min_num):

    df_in_copy = df_in.copy(deep=True)

    for col in df_in_copy.columns:
        count = (df_in_copy[col] >= min_value).sum()
        if count < min_value_min_num:
            print(count)
            df_in_copy.drop(col, axis=1, inplace=True)

    return df_in_copy


df = pd.DataFrame({
    'col1': [1, 1, 1, 1, 1, 1, 0],
    'col2': [1, 1, 1, 1, 1, 0, 0],
    'col3': [1, 1, 1, 1, 0, 0, 0],
    'col4': [1, 1, 1, 0, 0, 0, 0],
    'col5': [1, 1, 0, 0, 0, 0, 0],
    'col6': [1, 0, 0, 0, 0, 0, 0],
    'col7': [0, 0, 0, 0, 0, 0, 0]
})

min_value           = 1
min_value_min_num   = 5
df_filtered         = filter_col(df, min_value, min_value_min_num)


'''
print(df)

   col1  col2  col3  col4  col5  col6  col7
0     1     1     1     1     1     1     0
1     1     1     1     1     1     0     0
2     1     1     1     1     0     0     0
3     1     1     1     0     0     0     0
4     1     1     0     0     0     0     0
5     1     0     0     0     0     0     0
6     0     0     0     0     0     0     0

print(df_filtered)

   col1  col2
0     1     1
1     1     1
2     1     1
3     1     1
4     1     1
5     1     0
6     0     0
'''

######################### get_malthusian_selection_coefficients ##########################

# Equation for calculating mij in reference:
# Fitness effects of advantageous mutations in evolving Escherichia coli populations

import numpy as np


def get_malthusian_selection_coefficients(mutant_freq_start, mutant_freq_end, generation_num):

    malthusian_selection_coefficients = np.log((mutant_freq_end / mutant_freq_start) / ((1 - mutant_freq_end) / (1 - mutant_freq_start))) / generation_num

    return malthusian_selection_coefficients


mutant_freq_start   = 0.2
mutant_freq_end     = 0.8
generation_num      = 50
Malthusian_Selection_Coefficients = get_malthusian_selection_coefficients(mutant_freq_start, mutant_freq_end, generation_num)
print(Malthusian_Selection_Coefficients)


##################################### get percentile #####################################

import numpy as np

num_list = [1,2,3,4,5,6,7,8,9,10]
num_array = np.array(num_list)

percentile_25 = np.percentile(num_array, 25)
percentile_75 = np.percentile(num_array, 75)

print(percentile_25)	# 3.25
print(percentile_75)	# 7.75


###################################### transpose_csv #####################################

import pandas as pd


def transpose_csv(file_in, file_out, sep_symbol, column_name_pos, row_name_pos):

    csv = pd.read_csv(file_in, sep=sep_symbol, header=column_name_pos, index_col=row_name_pos)
    df_csv = pd.DataFrame(data=csv)
    transposed_csv = df_csv.T
    transposed_csv.to_csv(file_out, sep=sep_symbol)


file_in =  '/Users/songweizhi/Desktop/test.csv'
file_out = '/Users/songweizhi/Desktop/test_T.csv'
sep_symbol = '\t'    # ',' or '\t'
column_name_pos = 0  # set first row as column name
row_name_pos = 0     # set first column as row name
transpose_csv(file_in, file_out, sep_symbol, column_name_pos, row_name_pos)



################################## re-index column order #################################

import pandas

# turn list into arrary
category_num_arrary = np.array([[1,2,3], [4,5,6], [7,8,9]])

# add row and column name to dataframe
category_num_df = pandas.DataFrame(category_num_arrary, index=['Sample_1', 'Sample_2', 'Sample_3'], columns=['B', 'A', 'C'])
print(category_num_df)
# re-index column order
category_num_df = category_num_df.reindex(['A', 'B', 'C'], axis=1)
print(category_num_df)

# write out
#category_num_df.to_csv('Output.csv')

'''
          B  A  C                    A  B  C
Sample_1  1  2  3          Sample_1  2  1  3
Sample_2  4  5  6    -->   Sample_2  5  4  6
Sample_3  7  8  9          Sample_3  8  7  9

'''

##########################################################################################

import os
import numpy as np
import pandas as pd
from scipy import stats


def transpose_csv(file_in, file_out, sep_symbol, column_name_pos, row_name_pos):
    csv = pd.read_csv(file_in, sep=sep_symbol, header=column_name_pos, index_col=row_name_pos)
    df_csv = pd.DataFrame(data=csv)
    transposed_csv = df_csv.T
    transposed_csv.to_csv(file_out, sep=sep_symbol)


##########################################################################################

import math
print(int(math.ceil(11/float(3))))    # 4

# The divmod() returns a pair of numbers (a tuple) consisting of quotient q and remainder r
print(divmod(20, 3))    # (6, 2)


import numpy as np
from scipy import stats

a = [1, 2, 3, 3, 3, 3, 4, 5]
b = [5, 6, 7, 4, 6, 7, 4, 6, 6, 1]

print(np.sum(a))
print(np.mean(a))
print(np.std(a))
print('\n')
print(np.sum(b))
print(np.mean(b))
print(np.std(b))
print('\n')
p = stats.ttest_ind(a,b)
print(p)


##########################################################################################

note = '''

# ln in numpy: np.log()
print(np.log(2.7182780911361077))
# 0.9999986251147811

'''




########################################################################################################################

import os
import multiprocessing as mp


def exe_cmds(cmd_file, num_threads):

    cmd_set = set()
    for each_cmd in open(cmd_file):
        cmd_set.add(each_cmd.strip())
    
    print('Running %s commands with %s cores' % (len(cmd_set), num_threads))
    pool = mp.Pool(processes=num_threads)
    pool.map(os.system, cmd_set)
    pool.close()
    pool.join()
    print('Done!')


########################################################################################################################

import os
import glob
from Bio import SeqIO
import multiprocessing as mp


def sep_path_basename_ext(file_in):

    # separate path and file name
    f_path, file_name = os.path.split(file_in)
    if f_path == '':
        f_path = '.'

    # separate file basename and extension
    f_base, f_ext = os.path.splitext(file_name)

    return f_path, f_base, f_ext


def gbk2faa_worker(arg_list):

    gbk_in  = arg_list[0]
    faa_out = arg_list[1]

    faa_out_handle = open(faa_out, 'w')
    for seq_record in SeqIO.parse(gbk_in, 'genbank'):
        for feature in seq_record.features:
            if feature.type == 'CDS':
                feature_locus_tag = feature.qualifiers['locus_tag'][0]
                feature_translation = feature.qualifiers['translation'][0]
                faa_out_handle.write('>%s\n' % feature_locus_tag)
                faa_out_handle.write('%s\n' % feature_translation)
    faa_out_handle.close()


def gbk2faa(gbk_dir, gbk_ext, faa_dir, faa_ext, num_threads):

    gbk_file_re   = '%s/*.%s' % (gbk_dir, gbk_ext)
    gbk_file_list = glob.glob(gbk_file_re)

    list_of_arg_list = []
    for each_gbk in gbk_file_list:
        gbk_path, gbk_base, gbk_ext = sep_path_basename_ext(each_gbk)
        pwd_faa_out = '%s/%s.%s' % (faa_dir, gbk_base, faa_ext)
        list_of_arg_list.append([each_gbk, pwd_faa_out])

    print('Processing %s files with %s cores' % (len(list_of_arg_list), num_threads))
    pool = mp.Pool(processes=num_threads)
    pool.map(gbk2faa_worker, list_of_arg_list)
    pool.close()
    pool.join()
    print('Done')


gbk_folder          = '/Users/songweizhi/Desktop/gbk_files'
gbk_file_ext        = 'gbk'
output_faa_folder   = '/Users/songweizhi/Desktop/faa_files'
faa_file_ext        = 'faa'
num_threads         = 6
gbk2faa(gbk_folder, gbk_file_ext, output_faa_folder, faa_file_ext, num_threads)






file_in = 'input_file.txt'

col_index = {}
for each_line in open(file_in):
    each_line_split = each_line.strip().split('\t')
    if each_line.startswith('\t'):
        col_index = {key: i for i, key in enumerate(each_line_split)}
    else:
        gene_1    = each_line_split[col_index['gene_1']]
        gene_2    = each_line_split[col_index['gene_2']]
        direction = each_line_split[col_index['direction']]





######################## get file list ########################

import glob

file_dir = '/Users/songweizhi/Desktop/DateArTree'
file_ext = 'txt'

file_re = '%s/*.%s' % (file_dir, file_ext)
file_list = glob.glob(file_re)
print(file_list)


####################### get folder list #######################

import os

dir = '/Users/songweizhi/Desktop/DateArTree/0_HGT_ALE/op_qualified_OGs_gene_tree_dir'
sub_dir_list = next(os.walk(dir))[1]
print(sub_dir_list)


############# get_time_since_last_modification #############

import os
import time

def get_time_since_last_modification(target_folder_or_file):

    last_modified_time = os.path.getmtime(target_folder_or_file)
    current_time = time.time()

    tslm_sec  = current_time - last_modified_time
    tslm_min  = tslm_sec / 60
    tslm_hour = tslm_sec / (60 * 60)
    tslm_day  = tslm_sec / (60 * 60 * 24)

    tslm_sec  = float("{0:.2f}".format(tslm_sec))
    tslm_min  = float("{0:.2f}".format(tslm_min))
    tslm_hour = float("{0:.2f}".format(tslm_hour))
    tslm_day  = float("{0:.2f}".format(tslm_day))

    return tslm_sec, tslm_min, tslm_hour, tslm_day


target_file = '/Users/songweizhi/Desktop/foo.txt'
tslm_sec, tslm_min, tslm_hour, tslm_day = get_time_since_last_modification(target_file)


#################### whether folder/file exist #####################

import os

folder_1 = '/Users/songweizhi/Desktop/folder_1'
file_1 = '/Users/songweizhi/Desktop/folder_1/file.txt'

print(os.path.isdir(folder_1))
print(os.path.isfile(file_1))


######################## get file size ########################

text_file = '/Users/songweizhi/Desktop/a.txt'
text_file_size = os.stat(text_file).st_size


###############################################################

# Sometimes when we only care about certain parts of a tuple when unpacking, a dummy variable like _ is used as placeholder
___, filename = os.path.split('/home/luciano/.ssh/idrsa.pub')
print(filename)     # idrsa.pub


#################### remove folders or files #####################

def rm_folder_file(target_re):

    target_list = glob.glob(target_re)

    for target in target_list:

        if os.path.isdir(target) is True:
            os.system('rm -r %s' % target)
            
        elif os.path.isfile(target) is True:
            os.system('rm %s' % target)

